@@ -83,6 +83,8 @@ def set_default(key, value):
 
 
 def get_boolean(value):
+    if value is None:
+        return False
     if type(value) is bool:
         return value
     if value.lower() == 'true':
@@ -103,12 +105,11 @@ def _get_os_environ(envvar_prefix=None):
     for k in environkeys:
         newkey = k.split('_', 1)[-1].lower()
         ret[newkey] = os.environ[k]
-    # If the only environ key is region name, don't make a cloud, because
-    # it's being used as a cloud selector
-    if not environkeys or (
-            len(environkeys) == 1 and 'region_name' in ret):
-        return None
-    return ret
+    # If the only environ keys are cloud and region_name, don't return anything
+    # because they are cloud selectors
+    if set(environkeys) - set(['OS_CLOUD', 'OS_REGION_NAME']):
+        return ret
+    return None
 
 
 def _merge_clouds(old_dict, new_dict):
@@ -171,8 +172,10 @@ class OpenStackConfig(object):
 
     def __init__(self, config_files=None, vendor_files=None,
                  override_defaults=None, force_ipv4=None,
-                 envvar_prefix=None, secure_files=None):
+                 envvar_prefix=None, secure_files=None,
+                 pw_func=None, session_constructor=None):
         self.log = _log.setup_logging(__name__)
+        self._session_constructor = session_constructor
 
         self._config_files = config_files or CONFIG_FILES
         self._secure_files = secure_files or SECURE_FILES
@@ -286,6 +289,10 @@ class OpenStackConfig(object):
         # Flag location to hold the peeked value of an argparse timeout value
         self._argv_timeout = False
 
+        # Save the password callback
+        # password = self._pw_callback(prompt="Password: ")
+        self._pw_callback = pw_func
+
     def get_extra_config(self, key, defaults=None):
         """Fetch an arbitrary extra chunk of config, laying in defaults.
 
@@ -381,7 +388,10 @@ class OpenStackConfig(object):
         if 'regions' in config:
             return self._expand_regions(config['regions'])
         elif 'region_name' in config:
-            regions = config['region_name'].split(',')
+            if isinstance(config['region_name'], list):
+                regions = config['region_name']
+            else:
+                regions = config['region_name'].split(',')
             if len(regions) > 1:
                 warnings.warn(
                     "Comma separated lists in region_name are deprecated."
@@ -432,7 +442,7 @@ class OpenStackConfig(object):
         # Only validate cloud name if one was given
         if name and name not in self.cloud_config['clouds']:
             raise exceptions.OpenStackConfigException(
-                "Named cloud {name} requested that was not found.".format(
+                "Cloud {name} was not found.".format(
                     name=name))
 
         our_cloud = self.cloud_config['clouds'].get(name, dict())
@@ -466,6 +476,17 @@ class OpenStackConfig(object):
             else:
                 profile_data = vendors.get_profile(profile_name)
                 if profile_data:
+                    status = profile_data.pop('status', 'active')
+                    message = profile_data.pop('message', '')
+                    if status == 'deprecated':
+                        warnings.warn(
+                            "{profile_name} is deprecated: {message}".format(
+                                profile_name=profile_name, message=message))
+                    elif status == 'shutdown':
+                        raise exceptions.OpenStackConfigException(
+                            "{profile_name} references a cloud that no longer"
+                            " exists: {message}".format(
+                                profile_name=profile_name, message=message))
                     _auth_update(cloud, profile_data)
                 else:
                     # Can't find the requested vendor config, go about business
@@ -473,11 +494,74 @@ class OpenStackConfig(object):
                                   " the cloud '{1}'".format(profile_name,
                                                             name))
 
-    def _fix_backwards_madness(self, cloud):
-        cloud = self._fix_backwards_project(cloud)
-        cloud = self._fix_backwards_auth_plugin(cloud)
-        cloud = self._fix_backwards_interface(cloud)
-        cloud = self._handle_domain_id(cloud)
+    def _project_scoped(self, cloud):
+        return ('project_id' in cloud or 'project_name' in cloud
+                or 'project_id' in cloud['auth']
+                or 'project_name' in cloud['auth'])
+
+    def _validate_networks(self, networks, key):
+        value = None
+        for net in networks:
+            if value and net[key]:
+                raise exceptions.OpenStackConfigException(
+                    "Duplicate network entries for {key}: {net1} and {net2}."
+                    " Only one network can be flagged with {key}".format(
+                        key=key,
+                        net1=value['name'],
+                        net2=net['name']))
+            if not value and net[key]:
+                value = net
+
+    def _fix_backwards_networks(self, cloud):
+        # Leave the external_network and internal_network keys in the
+        # dict because consuming code might be expecting them.
+        networks = []
+        # Normalize existing network entries
+        for net in cloud.get('networks', []):
+            name = net.get('name')
+            if not name:
+                raise exceptions.OpenStackConfigException(
+                    'Entry in network list is missing required field "name".')
+            network = dict(
+                name=name,
+                routes_externally=get_boolean(net.get('routes_externally')),
+                nat_destination=get_boolean(net.get('nat_destination')),
+                default_interface=get_boolean(net.get('default_interface')),
+            )
+            # routes_ipv4_externally defaults to the value of routes_externally
+            network['routes_ipv4_externally'] = get_boolean(
+                net.get(
+                    'routes_ipv4_externally', network['routes_externally']))
+            # routes_ipv6_externally defaults to the value of routes_externally
+            network['routes_ipv6_externally'] = get_boolean(
+                net.get(
+                    'routes_ipv6_externally', network['routes_externally']))
+            networks.append(network)
+
+        for key in ('external_network', 'internal_network'):
+            external = key.startswith('external')
+            if key in cloud and 'networks' in cloud:
+                raise exceptions.OpenStackConfigException(
+                    "Both {key} and networks were specified in the config."
+                    " Please remove {key} from the config and use the network"
+                    " list to configure network behavior.".format(key=key))
+            if key in cloud:
+                warnings.warn(
+                    "{key} is deprecated. Please replace with an entry in"
+                    " a dict inside of the networks list with name: {name}"
+                    " and routes_externally: {external}".format(
+                        key=key, name=cloud[key], external=external))
+                networks.append(dict(
+                    name=cloud[key],
+                    routes_externally=external,
+                    nat_destination=not external,
+                    default_interface=external))
+
+        # Validate that we don't have duplicates
+        self._validate_networks(networks, 'nat_destination')
+        self._validate_networks(networks, 'default_interface')
+
+        cloud['networks'] = networks
         return cloud
 
     def _handle_domain_id(self, cloud):
@@ -487,9 +571,14 @@ class OpenStackConfig(object):
             'domain_name': ('user_domain_name', 'project_domain_name'),
         }
         for target_key, possible_values in mappings.items():
+            if not self._project_scoped(cloud):
+                if target_key in cloud and target_key not in cloud['auth']:
+                    cloud['auth'][target_key] = cloud.pop(target_key)
+                continue
             for key in possible_values:
                 if target_key in cloud['auth'] and key not in cloud['auth']:
                     cloud['auth'][key] = cloud['auth'][target_key]
+            cloud.pop(target_key, None)
             cloud['auth'].pop(target_key, None)
         return cloud
 
@@ -498,10 +587,6 @@ class OpenStackConfig(object):
         # Also handle moving domain names into auth so that domain mapping
         # is easier
         mappings = {
-            'project_id': ('tenant_id', 'tenant-id',
-                           'project_id', 'project-id'),
-            'project_name': ('tenant_name', 'tenant-name',
-                             'project_name', 'project-name'),
             'domain_id': ('domain_id', 'domain-id'),
             'domain_name': ('domain_name', 'domain-name'),
             'user_domain_id': ('user_domain_id', 'user-domain-id'),
@@ -511,15 +596,32 @@ class OpenStackConfig(object):
                 'project_domain_name', 'project-domain-name'),
             'token': ('auth-token', 'auth_token', 'token'),
         }
+        if cloud.get('auth_type', None) == 'v2password':
+            # If v2password is explcitly requested, this is to deal with old
+            # clouds. That's fine - we need to map settings in the opposite
+            # direction
+            mappings['tenant_id'] = (
+                'project_id', 'project-id', 'tenant_id', 'tenant-id')
+            mappings['tenant_name'] = (
+                'project_name', 'project-name', 'tenant_name', 'tenant-name')
+        else:
+            mappings['project_id'] = (
+                'tenant_id', 'tenant-id', 'project_id', 'project-id')
+            mappings['project_name'] = (
+                'tenant_name', 'tenant-name', 'project_name', 'project-name')
+        # Special-case username and password so that we don't have to load
+        # the plugins so early
+        mappings['username'] = ('user-name', 'user_name', 'username')
+        mappings['password'] = ('password',)
         for target_key, possible_values in mappings.items():
             target = None
             for key in possible_values:
-                if key in cloud:
-                    target = str(cloud[key])
-                    del cloud[key]
-                if key in cloud['auth']:
-                    target = str(cloud['auth'][key])
-                    del cloud['auth'][key]
+                root_target = cloud.pop(key, None)
+                auth_target = cloud['auth'].pop(key, None)
+                if root_target:
+                    target = str(root_target)
+                elif auth_target:
+                    target = str(auth_target)
             if target:
                 cloud['auth'][target_key] = target
         return cloud
@@ -540,11 +642,9 @@ class OpenStackConfig(object):
         # use of the auth plugin that can do auto-selection and dealing
         # with that based on auth parameters. v2password is basically
         # completely broken
-        if cloud['auth_type'] == 'v2password':
-            cloud['auth_type'] = 'password'
         return cloud
 
-    def register_argparse_arguments(self, parser, argv, service_keys=[]):
+    def register_argparse_arguments(self, parser, argv, service_keys=None):
         """Register all of the common argparse options needed.
 
         Given an argparse parser, register the keystoneauth Session arguments,
@@ -563,6 +663,9 @@ class OpenStackConfig(object):
                                                     is requested
         """
 
+        if service_keys is None:
+            service_keys = []
+
         # Fix argv in place - mapping any keys with embedded _ in them to -
         _fix_argv(argv)
 
@@ -687,7 +790,7 @@ class OpenStackConfig(object):
                         cloud, region_name=region['name']))
         return clouds
 
-    def _fix_args(self, args, argparse=None):
+    def _fix_args(self, args=None, argparse=None):
         """Massage the passed-in options
 
         Replace - with _ and strip os_ prefixes.
@@ -695,6 +798,8 @@ class OpenStackConfig(object):
         Convert an argparse Namespace object to a dict, removing values
         that are either None or ''.
         """
+        if not args:
+            args = {}
 
         if argparse:
             # Convert the passed-in Namespace
@@ -736,7 +841,7 @@ class OpenStackConfig(object):
     def auth_config_hook(self, config):
         """Allow examination of config values before loading auth plugin
 
-        OpenStackClient will override this to perform additional chacks
+        OpenStackClient will override this to perform additional checks
         on auth_type.
         """
         return config
@@ -753,56 +858,13 @@ class OpenStackConfig(object):
             # Set to notused rather than None because validate_auth will
             # strip the value if it's actually python None
             config['auth']['token'] = 'notused'
+        elif config['auth_type'] == 'token_endpoint':
+            # Humans have been trained to use a thing called token_endpoint
+            # That it does not exist in keystoneauth is irrelvant- it not
+            # doing what they want causes them sorrow.
+            config['auth_type'] = 'admin_token'
         return loading.get_plugin_loader(config['auth_type'])
 
-    def _validate_auth_ksc(self, config):
-        try:
-            import keystoneclient.auth as ksc_auth
-        except ImportError:
-            return config
-
-        # May throw a keystoneclient.exceptions.NoMatchingPlugin
-        plugin_options = ksc_auth.get_plugin_class(
-            config['auth_type']).get_options()
-
-        for p_opt in plugin_options:
-            # if it's in config.auth, win, kill it from config dict
-            # if it's in config and not in config.auth, move it
-            # deprecated loses to current
-            # provided beats default, deprecated or not
-            winning_value = self._find_winning_auth_value(
-                p_opt, config['auth'])
-            if not winning_value:
-                winning_value = self._find_winning_auth_value(p_opt, config)
-
-            # if the plugin tells us that this value is required
-            # then error if it's doesn't exist now
-            if not winning_value and p_opt.required:
-                raise exceptions.OpenStackConfigException(
-                    'Unable to find auth information for cloud'
-                    ' {cloud} in config files {files}'
-                    ' or environment variables. Missing value {auth_key}'
-                    ' required for auth plugin {plugin}'.format(
-                        cloud=cloud, files=','.join(self._config_files),
-                        auth_key=p_opt.name, plugin=config.get('auth_type')))
-
-            # Clean up after ourselves
-            for opt in [p_opt.name] + [o.name for o in p_opt.deprecated_opts]:
-                opt = opt.replace('-', '_')
-                config.pop(opt, None)
-                config['auth'].pop(opt, None)
-
-            if winning_value:
-                # Prefer the plugin configuration dest value if the value's key
-                # is marked as depreciated.
-                if p_opt.dest is None:
-                    config['auth'][p_opt.name.replace('-', '_')] = (
-                        winning_value)
-                else:
-                    config['auth'][p_opt.dest] = winning_value
-
-        return config
-
     def _validate_auth(self, config, loader):
         # May throw a keystoneauth1.exceptions.NoMatchingPlugin
 
@@ -814,25 +876,128 @@ class OpenStackConfig(object):
             # deprecated loses to current
             # provided beats default, deprecated or not
             winning_value = self._find_winning_auth_value(
-                p_opt, config['auth'])
+                p_opt,
+                config['auth'],
+            )
             if not winning_value:
-                winning_value = self._find_winning_auth_value(p_opt, config)
+                winning_value = self._find_winning_auth_value(
+                    p_opt,
+                    config,
+                )
 
-            # Clean up after ourselves
-            for opt in [p_opt.name] + [o.name for o in p_opt.deprecated]:
-                opt = opt.replace('-', '_')
-                config.pop(opt, None)
-                config['auth'].pop(opt, None)
+            config = self._clean_up_after_ourselves(
+                config,
+                p_opt,
+                winning_value,
+            )
 
             if winning_value:
                 # Prefer the plugin configuration dest value if the value's key
-                # is marked as depreciated.
+                # is marked as deprecated.
                 if p_opt.dest is None:
-                    config['auth'][p_opt.name.replace('-', '_')] = (
-                        winning_value)
+                    good_name = p_opt.name.replace('-', '_')
+                    config['auth'][good_name] = winning_value
                 else:
                     config['auth'][p_opt.dest] = winning_value
 
+            # See if this needs a prompting
+            config = self.option_prompt(config, p_opt)
+
+        return config
+
+    def _validate_auth_correctly(self, config, loader):
+        # May throw a keystoneauth1.exceptions.NoMatchingPlugin
+
+        plugin_options = loader.get_options()
+
+        for p_opt in plugin_options:
+            # if it's in config, win, move it and kill it from config dict
+            # if it's in config.auth but not in config it's good
+            # deprecated loses to current
+            # provided beats default, deprecated or not
+            winning_value = self._find_winning_auth_value(
+                p_opt,
+                config,
+            )
+            if not winning_value:
+                winning_value = self._find_winning_auth_value(
+                    p_opt,
+                    config['auth'],
+                )
+
+            config = self._clean_up_after_ourselves(
+                config,
+                p_opt,
+                winning_value,
+            )
+
+            # See if this needs a prompting
+            config = self.option_prompt(config, p_opt)
+
+        return config
+
+    def option_prompt(self, config, p_opt):
+        """Prompt user for option that requires a value"""
+        if (
+                p_opt.prompt is not None and
+                p_opt.dest not in config['auth'] and
+                self._pw_callback is not None
+        ):
+            config['auth'][p_opt.dest] = self._pw_callback(p_opt.prompt)
+        return config
+
+    def _clean_up_after_ourselves(self, config, p_opt, winning_value):
+
+        # Clean up after ourselves
+        for opt in [p_opt.name] + [o.name for o in p_opt.deprecated]:
+            opt = opt.replace('-', '_')
+            config.pop(opt, None)
+            config['auth'].pop(opt, None)
+
+        if winning_value:
+            # Prefer the plugin configuration dest value if the value's key
+            # is marked as depreciated.
+            if p_opt.dest is None:
+                config['auth'][p_opt.name.replace('-', '_')] = (
+                    winning_value)
+            else:
+                config['auth'][p_opt.dest] = winning_value
+        return config
+
+    def magic_fixes(self, config):
+        """Perform the set of magic argument fixups"""
+
+        # Infer token plugin if a token was given
+        if (('auth' in config and 'token' in config['auth']) or
+                ('auth_token' in config and config['auth_token']) or
+                ('token' in config and config['token'])):
+            config.setdefault('token', config.pop('auth_token', None))
+            config.setdefault('auth_type', 'token')
+
+        # These backwards compat values are only set via argparse. If it's
+        # there, it's because it was passed in explicitly, and should win
+        config = self._fix_backwards_api_timeout(config)
+        if 'endpoint_type' in config:
+            config['interface'] = config.pop('endpoint_type')
+
+        config = self._fix_backwards_auth_plugin(config)
+        config = self._fix_backwards_project(config)
+        config = self._fix_backwards_interface(config)
+        config = self._fix_backwards_networks(config)
+        config = self._handle_domain_id(config)
+
+        for key in BOOL_KEYS:
+            if key in config:
+                if type(config[key]) is not bool:
+                    config[key] = get_boolean(config[key])
+
+        # TODO(mordred): Special casing auth_url here. We should
+        #                come back to this betterer later so that it's
+        #                more generalized
+        if 'auth' in config and 'auth_url' in config['auth']:
+            config['auth']['auth_url'] = config['auth']['auth_url'].format(
+                **config)
+
         return config
 
     def get_one_cloud(self, cloud=None, validate=True,
@@ -851,6 +1016,111 @@ class OpenStackConfig(object):
         :param region_name: Name of the region of the cloud.
         :param kwargs: Additional configuration options
 
+        :raises: keystoneauth1.exceptions.MissingRequiredOptions
+            on missing required auth parameters
+        """
+        args = self._fix_args(kwargs, argparse=argparse)
+
+        if cloud is None:
+            if 'cloud' in args:
+                cloud = args['cloud']
+            else:
+                cloud = self.default_cloud
+
+        config = self._get_base_cloud_config(cloud)
+
+        # Get region specific settings
+        if 'region_name' not in args:
+            args['region_name'] = ''
+        region = self._get_region(cloud=cloud, region_name=args['region_name'])
+        args['region_name'] = region['name']
+        region_args = copy.deepcopy(region['values'])
+
+        # Regions is a list that we can use to create a list of cloud/region
+        # objects. It does not belong in the single-cloud dict
+        config.pop('regions', None)
+
+        # Can't just do update, because None values take over
+        for arg_list in region_args, args:
+            for (key, val) in iter(arg_list.items()):
+                if val is not None:
+                    if key == 'auth' and config[key] is not None:
+                        config[key] = _auth_update(config[key], val)
+                    else:
+                        config[key] = val
+
+        config = self.magic_fixes(config)
+        config = self._normalize_keys(config)
+
+        # NOTE(dtroyer): OSC needs a hook into the auth args before the
+        #                plugin is loaded in order to maintain backward-
+        #                compatible behaviour
+        config = self.auth_config_hook(config)
+
+        if validate:
+            try:
+                loader = self._get_auth_loader(config)
+                config = self._validate_auth(config, loader)
+                auth_plugin = loader.load_from_options(**config['auth'])
+            except Exception as e:
+                # We WANT the ksa exception normally
+                # but OSC can't handle it right now, so we try deferring
+                # to ksc. If that ALSO fails, it means there is likely
+                # a deeper issue, so we assume the ksa error was correct
+                self.log.debug("Deferring keystone exception: {e}".format(e=e))
+                auth_plugin = None
+                try:
+                    config = self._validate_auth_ksc(config, cloud)
+                except Exception:
+                    raise e
+        else:
+            auth_plugin = None
+
+        # If any of the defaults reference other values, we need to expand
+        for (key, value) in config.items():
+            if hasattr(value, 'format'):
+                config[key] = value.format(**config)
+
+        force_ipv4 = config.pop('force_ipv4', self.force_ipv4)
+        prefer_ipv6 = config.pop('prefer_ipv6', True)
+        if not prefer_ipv6:
+            force_ipv4 = True
+
+        if cloud is None:
+            cloud_name = ''
+        else:
+            cloud_name = str(cloud)
+        return cloud_config.CloudConfig(
+            name=cloud_name,
+            region=config['region_name'],
+            config=config,
+            force_ipv4=force_ipv4,
+            auth_plugin=auth_plugin,
+            openstack_config=self,
+            session_constructor=self._session_constructor,
+        )
+
+    def get_one_cloud_osc(
+        self,
+        cloud=None,
+        validate=True,
+        argparse=None,
+        **kwargs
+    ):
+        """Retrieve a single cloud configuration and merge additional options
+
+        :param string cloud:
+            The name of the configuration to load from clouds.yaml
+        :param boolean validate:
+            Validate the config. Setting this to False causes no auth plugin
+            to be created. It's really only useful for testing.
+        :param Namespace argparse:
+            An argparse Namespace object; allows direct passing in of
+            argparse options to be added to the cloud config.  Values
+            of None and '' will be removed.
+        :param region_name: Name of the region of the cloud.
+        :param kwargs: Additional configuration options
+
         :raises: keystoneauth1.exceptions.MissingRequiredOptions
             on missing required auth parameters
         """
@@ -885,31 +1155,7 @@ class OpenStackConfig(object):
                     else:
                         config[key] = val
 
-        # Infer token plugin if a token was given
-        if (('auth' in config and 'token' in config['auth']) or
-                ('auth_token' in config and config['auth_token']) or
-                ('token' in config and config['token'])):
-            config.setdefault('token', config.pop('auth_token', None))
-
-        # These backwards compat values are only set via argparse. If it's
-        # there, it's because it was passed in explicitly, and should win
-        config = self._fix_backwards_api_timeout(config)
-        if 'endpoint_type' in config:
-            config['interface'] = config.pop('endpoint_type')
-
-        config = self._fix_backwards_madness(config)
-
-        for key in BOOL_KEYS:
-            if key in config:
-                if type(config[key]) is not bool:
-                    config[key] = get_boolean(config[key])
-
-        # TODO(mordred): Special casing auth_url here. We should
-        #                come back to this betterer later so that it's
-        #                more generalized
-        if 'auth' in config and 'auth_url' in config['auth']:
-            config['auth']['auth_url'] = config['auth']['auth_url'].format(
-                **config)
+        config = self.magic_fixes(config)
 
         # NOTE(dtroyer): OSC needs a hook into the auth args before the
         #                plugin is loaded in order to maintain backward-
@@ -917,21 +1163,9 @@ class OpenStackConfig(object):
         config = self.auth_config_hook(config)
 
         if validate:
-            try:
-                loader = self._get_auth_loader(config)
-                config = self._validate_auth(config, loader)
-                auth_plugin = loader.load_from_options(**config['auth'])
-            except Exception as e:
-                # We WANT the ksa exception normally
-                # but OSC can't handle it right now, so we try deferring
-                # to ksc. If that ALSO fails, it means there is likely
-                # a deeper issue, so we assume the ksa error was correct
-                self.log.debug("Deferring keystone exception: {e}".format(e=e))
-                auth_plugin = None
-                try:
-                    config = self._validate_auth_ksc(config)
-                except Exception:
-                    raise e
+            loader = self._get_auth_loader(config)
+            config = self._validate_auth_correctly(config, loader)
+            auth_plugin = loader.load_from_options(**config['auth'])
         else:
             auth_plugin = None
 
@@ -950,12 +1184,13 @@ class OpenStackConfig(object):
         else:
             cloud_name = str(cloud)
         return cloud_config.CloudConfig(
-            name=cloud_name, region=config['region_name'],
+            name=cloud_name,
+            region=config['region_name'],
             config=self._normalize_keys(config),
             force_ipv4=force_ipv4,
             auth_plugin=auth_plugin,
-            openstack_config=self
-            )
+            openstack_config=self,
+        )
 
     @staticmethod
     def set_one_cloud(config_file, cloud, set_config=None):
