@@ -27,6 +27,11 @@ from os_client_config import exceptions
 from os_client_config.tests import base
 
 
+def prompt_for_password(prompt=None):
+    """Fake prompt function that just returns a constant string"""
+    return 'promptpass'
+
+
 class TestConfig(base.TestCase):
 
     def test_get_all_clouds(self):
@@ -221,7 +226,7 @@ class TestConfig(base.TestCase):
         c = config.OpenStackConfig(config_files=['nonexistent'],
                                    vendor_files=['nonexistent'],
                                    secure_files=[self.secure_yaml])
-        cc = c.get_one_cloud(cloud='_test_cloud_no_vendor')
+        cc = c.get_one_cloud(cloud='_test_cloud_no_vendor', validate=False)
         self.assertEqual('testpass', cc.auth['password'])
 
     def test_get_cloud_names(self):
@@ -361,7 +366,6 @@ class TestConfigArgparse(base.TestCase):
             project_name='project',
             region_name='region2',
             snack_type='cookie',
-            os_auth_token='no-good-things',
         )
 
         self.options = argparse.Namespace(**self.args)
@@ -415,6 +419,42 @@ class TestConfigArgparse(base.TestCase):
         self.assertEqual(cc.auth['password'], 'argpass')
         self.assertEqual(cc.snack_type, 'cookie')
 
+    def test_get_one_cloud_precedence_osc(self):
+        c = config.OpenStackConfig(
+            config_files=[self.cloud_yaml],
+            vendor_files=[self.vendor_yaml],
+        )
+
+        kwargs = {
+            'auth': {
+                'username': 'testuser',
+                'password': 'authpass',
+                'project-id': 'testproject',
+                'auth_url': 'http://example.com/v2',
+            },
+            'region_name': 'kwarg_region',
+            'password': 'ansible_password',
+            'arbitrary': 'value',
+        }
+
+        args = dict(
+            auth_url='http://example.com/v2',
+            username='user',
+            password='argpass',
+            project_name='project',
+            region_name='region2',
+            snack_type='cookie',
+        )
+
+        options = argparse.Namespace(**args)
+        cc = c.get_one_cloud_osc(
+            argparse=options,
+            **kwargs
+        )
+        self.assertEqual(cc.region_name, 'region2')
+        self.assertEqual(cc.auth['password'], 'argpass')
+        self.assertEqual(cc.snack_type, 'cookie')
+
     def test_get_one_cloud_precedence_no_argparse(self):
         c = config.OpenStackConfig(config_files=[self.cloud_yaml],
                                    vendor_files=[self.vendor_yaml])
@@ -433,7 +473,7 @@ class TestConfigArgparse(base.TestCase):
 
         cc = c.get_one_cloud(**kwargs)
         self.assertEqual(cc.region_name, 'kwarg_region')
-        self.assertEqual(cc.auth['password'], 'authpass')
+        self.assertEqual(cc.auth['password'], 'ansible_password')
         self.assertIsNone(cc.password)
 
     def test_get_one_cloud_just_argparse(self):
@@ -608,11 +648,30 @@ class TestConfigArgparse(base.TestCase):
         parser.add_argument('--os-auth-token')
         opts, _remain = parser.parse_known_args(
             ['--os-auth-token', 'very-bad-things',
-             '--os-auth-type', 'token'])
+             '--os-auth-type', 'token',
+             '--os-auth-url', 'http://example.com/v2',
+             '--os-project-name', 'project'])
         cc = c.get_one_cloud(argparse=opts)
         self.assertEqual(cc.config['auth_type'], 'token')
         self.assertEqual(cc.config['auth']['token'], 'very-bad-things')
 
+    def test_argparse_username_token(self):
+        c = config.OpenStackConfig(config_files=[self.cloud_yaml],
+                                   vendor_files=[self.vendor_yaml])
+
+        parser = argparse.ArgumentParser()
+        c.register_argparse_arguments(parser, [])
+        # novaclient will add this
+        parser.add_argument('--os-auth-token')
+        opts, _remain = parser.parse_known_args(
+            ['--os-auth-token', 'very-bad-things',
+             '--os-auth-type', 'token',
+             '--os-auth-url', 'http://example.com/v2',
+             '--os-username', 'user',
+             '--os-project-name', 'project'])
+        self.assertRaises(
+            TypeError, c.get_one_cloud, argparse=opts)
+
     def test_argparse_underscores(self):
         c = config.OpenStackConfig(config_files=[self.no_yaml],
                                    vendor_files=[self.no_yaml],
@@ -694,7 +753,7 @@ class TestConfigArgparse(base.TestCase):
         self.assertEqual(opts.http_timeout, '20')
         with testtools.ExpectedException(AttributeError):
             opts.os_network_service_type
-        cloud = c.get_one_cloud(argparse=opts, verify=False)
+        cloud = c.get_one_cloud(argparse=opts, validate=False)
         self.assertEqual(cloud.config['service_type'], 'network')
         self.assertEqual(cloud.config['interface'], 'admin')
         self.assertEqual(cloud.config['api_timeout'], '20')
@@ -712,10 +771,10 @@ class TestConfigArgparse(base.TestCase):
         opts, _remain = parser.parse_known_args(args)
         self.assertEqual(opts.os_service_type, 'network')
         self.assertEqual(opts.os_endpoint_type, 'admin')
-        self.assertEqual(opts.os_network_service_type, None)
-        self.assertEqual(opts.os_network_api_version, None)
+        self.assertIsNone(opts.os_network_service_type)
+        self.assertIsNone(opts.os_network_api_version)
         self.assertEqual(opts.network_api_version, '4')
-        cloud = c.get_one_cloud(argparse=opts, verify=False)
+        cloud = c.get_one_cloud(argparse=opts, validate=False)
         self.assertEqual(cloud.config['service_type'], 'network')
         self.assertEqual(cloud.config['interface'], 'admin')
         self.assertEqual(cloud.config['network_api_version'], '4')
@@ -735,14 +794,14 @@ class TestConfigArgparse(base.TestCase):
             parser, args, ['compute', 'network', 'volume'])
         opts, _remain = parser.parse_known_args(args)
         self.assertEqual(opts.os_network_service_type, 'badtype')
-        self.assertEqual(opts.os_compute_service_type, None)
-        self.assertEqual(opts.os_volume_service_type, None)
+        self.assertIsNone(opts.os_compute_service_type)
+        self.assertIsNone(opts.os_volume_service_type)
         self.assertEqual(opts.os_service_type, 'compute')
         self.assertEqual(opts.os_compute_service_name, 'cloudServers')
         self.assertEqual(opts.os_endpoint_type, 'admin')
-        self.assertEqual(opts.os_network_api_version, None)
+        self.assertIsNone(opts.os_network_api_version)
         self.assertEqual(opts.network_api_version, '4')
-        cloud = c.get_one_cloud(argparse=opts, verify=False)
+        cloud = c.get_one_cloud(argparse=opts, validate=False)
         self.assertEqual(cloud.config['service_type'], 'compute')
         self.assertEqual(cloud.config['network_service_type'], 'badtype')
         self.assertEqual(cloud.config['interface'], 'admin')
@@ -751,6 +810,38 @@ class TestConfigArgparse(base.TestCase):
         self.assertNotIn('http_timeout', cloud.config)
 
 
+class TestConfigPrompt(base.TestCase):
+
+    def setUp(self):
+        super(TestConfigPrompt, self).setUp()
+
+        self.args = dict(
+            auth_url='http://example.com/v2',
+            username='user',
+            project_name='project',
+            # region_name='region2',
+            auth_type='password',
+        )
+
+        self.options = argparse.Namespace(**self.args)
+
+    def test_get_one_cloud_prompt(self):
+        c = config.OpenStackConfig(
+            config_files=[self.cloud_yaml],
+            vendor_files=[self.vendor_yaml],
+            pw_func=prompt_for_password,
+        )
+
+        # This needs a cloud definition without a password.
+        # If this starts failing unexpectedly check that the cloud_yaml
+        # and/or vendor_yaml do not have a password in the selected cloud.
+        cc = c.get_one_cloud(
+            cloud='_test_cloud_no_vendor',
+            argparse=self.options,
+        )
+        self.assertEqual('promptpass', cc.auth['password'])
+
+
 class TestConfigDefault(base.TestCase):
 
     def setUp(self):
