@@ -19,6 +19,7 @@ from os_client_config import exceptions
 from os_client_config.tests import base
 
 import fixtures
+import keystoneauth1.exceptions
 
 
 class TestEnviron(base.TestCase):
@@ -29,6 +30,8 @@ class TestEnviron(base.TestCase):
             fixtures.EnvironmentVariable('OS_AUTH_URL', 'https://example.com'))
         self.useFixture(
             fixtures.EnvironmentVariable('OS_USERNAME', 'testuser'))
+        self.useFixture(
+            fixtures.EnvironmentVariable('OS_PASSWORD', 'testpass'))
         self.useFixture(
             fixtures.EnvironmentVariable('OS_PROJECT_NAME', 'testproject'))
         self.useFixture(
@@ -57,13 +60,15 @@ class TestEnviron(base.TestCase):
         self.useFixture(
             fixtures.EnvironmentVariable('OS_PREFER_IPV6', 'false'))
         c = config.OpenStackConfig(config_files=[self.cloud_yaml],
-                                   vendor_files=[self.vendor_yaml])
+                                   vendor_files=[self.vendor_yaml],
+                                   secure_files=[self.secure_yaml])
         cc = c.get_one_cloud('_test-cloud_')
         self.assertFalse(cc.prefer_ipv6)
 
     def test_environ_exists(self):
         c = config.OpenStackConfig(config_files=[self.cloud_yaml],
-                                   vendor_files=[self.vendor_yaml])
+                                   vendor_files=[self.vendor_yaml],
+                                   secure_files=[self.secure_yaml])
         cc = c.get_one_cloud('envvars')
         self._assert_cloud_details(cc)
         self.assertNotIn('auth_url', cc.config)
@@ -78,7 +83,8 @@ class TestEnviron(base.TestCase):
     def test_environ_prefix(self):
         c = config.OpenStackConfig(config_files=[self.cloud_yaml],
                                    vendor_files=[self.vendor_yaml],
-                                   envvar_prefix='NOVA_')
+                                   envvar_prefix='NOVA_',
+                                   secure_files=[self.secure_yaml])
         cc = c.get_one_cloud('envvars')
         self._assert_cloud_details(cc)
         self.assertNotIn('auth_url', cc.config)
@@ -92,7 +98,8 @@ class TestEnviron(base.TestCase):
 
     def test_get_one_cloud_with_config_files(self):
         c = config.OpenStackConfig(config_files=[self.cloud_yaml],
-                                   vendor_files=[self.vendor_yaml])
+                                   vendor_files=[self.vendor_yaml],
+                                   secure_files=[self.secure_yaml])
         self.assertIsInstance(c.cloud_config, dict)
         self.assertIn('cache', c.cloud_config)
         self.assertIsInstance(c.cloud_config['cache'], dict)
@@ -133,21 +140,45 @@ class TestEnvvars(base.TestCase):
         self.assertRaises(
             exceptions.OpenStackConfigException, c.get_one_cloud, 'envvars')
 
-    def test_have_envvars(self):
+    def test_incomplete_envvars(self):
         self.useFixture(
             fixtures.EnvironmentVariable('NOVA_USERNAME', 'nova'))
         self.useFixture(
             fixtures.EnvironmentVariable('OS_USERNAME', 'user'))
+        c = config.OpenStackConfig(
+            config_files=[self.cloud_yaml], vendor_files=[self.vendor_yaml])
+        self.assertRaises(
+            keystoneauth1.exceptions.auth_plugins.MissingRequiredOptions,
+            c.get_one_cloud, 'envvars')
+
+    def test_have_envvars(self):
+        self.useFixture(
+            fixtures.EnvironmentVariable('NOVA_USERNAME', 'nova'))
+        self.useFixture(
+            fixtures.EnvironmentVariable('OS_AUTH_URL', 'http://example.com'))
+        self.useFixture(
+            fixtures.EnvironmentVariable('OS_USERNAME', 'user'))
+        self.useFixture(
+            fixtures.EnvironmentVariable('OS_PASSWORD', 'password'))
+        self.useFixture(
+            fixtures.EnvironmentVariable('OS_PROJECT_NAME', 'project'))
         c = config.OpenStackConfig(config_files=[self.cloud_yaml],
                                    vendor_files=[self.vendor_yaml])
-        cc = c.get_one_cloud('envvars')
+        cc = c.get_one_cloud('envvars', validate=False)
         self.assertEqual(cc.config['auth']['username'], 'user')
 
     def test_old_envvars(self):
         self.useFixture(
             fixtures.EnvironmentVariable('NOVA_USERNAME', 'nova'))
+        self.useFixture(
+            fixtures.EnvironmentVariable(
+                'NOVA_AUTH_URL', 'http://example.com'))
+        self.useFixture(
+            fixtures.EnvironmentVariable('NOVA_PASSWORD', 'password'))
+        self.useFixture(
+            fixtures.EnvironmentVariable('NOVA_PROJECT_NAME', 'project'))
         c = config.OpenStackConfig(config_files=[self.cloud_yaml],
                                    vendor_files=[self.vendor_yaml],
                                    envvar_prefix='NOVA_')
-        cc = c.get_one_cloud('envvars')
+        cc = c.get_one_cloud('envvars', validate=False)
         self.assertEqual(cc.config['auth']['username'], 'nova')
